﻿using MindFusion.Diagramming.Wpf;
using MindFusion.Svg;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Reflection;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Xml;
using PixelFormat = System.Drawing.Imaging.PixelFormat;
using RenderOptions = MindFusion.Diagramming.Wpf.RenderOptions;

namespace MindfusionNewVersionIssues;

public static class ImageUtils
{
    private const PixelFormat StandartPixelFormat = PixelFormat.Format32bppArgb;
    private const int BitmapDpi = 240;
    private const float QualityModifier = 2.5f;

    private static SvgContent CreateSvgContent(string path)
    {
        using var fs = File.OpenRead(path);

        var xmlDocument = new XmlDocument();
        xmlDocument.Load(fs);

        var svgContent = new SvgContent();
        svgContent.Parse(xmlDocument, splitGroups: false);

        return svgContent;
    }

    public static BitmapSource CreateImage(string path)
    {
        using var fs = File.OpenRead(path);
        return CreateImage(fs);
    }

    private static ImageSource CreateImageFromSvg(SvgContent svgContent, Rect bounds,
        RenderOptions renderOptions)
    {
        var svgNode = new SvgNode
        {
            Bounds = bounds,
            Content = svgContent,
            Transparent = true,
        };

        var visual = new DrawingVisual();
        using var drawingContext = visual.RenderOpen();

        svgNode.DrawLocal(drawingContext, renderOptions);
        drawingContext.Close();

        var image = new RenderTargetBitmap(
            pixelWidth: Convert.ToInt32(bounds.Width * QualityModifier),
            pixelHeight: Convert.ToInt32(bounds.Height * QualityModifier),
            dpiX: BitmapDpi,
            dpiY: BitmapDpi,
            PixelFormats.Pbgra32);

        image.Render(visual);
        image.Freeze();

        // We need to force dispose render target bitmap after freezing it.
        Dispose(image);

        return image;
    }

    /// <summary>
    /// Force disposing of a render target bitmap.
    /// WPF Hack: Imaging - Force the RenderTargetBitmap to release memory handles immediately.
    /// https://github.com/dotnet/wpf/issues/3067.
    /// </summary>
    /// <param name="bitmap">Bitmap to dispose.</param>
    public static void Dispose(RenderTargetBitmap bitmap)
    {
        var flags = BindingFlags.Instance | BindingFlags.NonPublic;
        var fieldRenderTargetBitmap = typeof(RenderTargetBitmap).GetField("_renderTargetBitmap", flags);
        var renderTargetBitmap = fieldRenderTargetBitmap.GetValue(bitmap) as IDisposable;
        renderTargetBitmap?.Dispose();
    }

    private static BitmapSource CreateImage(Stream imageStream)
    {
        const int MaximumImageSize = 2048;

        if (imageStream.Length == 0)
        {
            return default;
        }

        if (imageStream.CanSeek)
        {
            imageStream.Seek(0, SeekOrigin.Begin);
        }

        using var bitmap = new Bitmap(imageStream);
        bitmap.SetResolution(240, 240);

        // This logic allows to adjust pixel format of bitmap.
        if (bitmap.PixelFormat != StandartPixelFormat)
        {
            return CreateBitmapSourceInternal(bitmap);
        }

        var memoryStream = new MemoryStream();
        var bitmapImage = new BitmapImage();

        bitmap.Save(memoryStream, bitmap.RawFormat);
        bitmapImage.BeginInit();

        bitmapImage.CacheOption = BitmapCacheOption.OnLoad;
        bitmapImage.StreamSource = memoryStream;

        // Adjust pixel height and width if it is higher that maximum
        // to optimize memory usage.
        if (bitmap.Height > MaximumImageSize)
        {
            bitmapImage.DecodePixelHeight = MaximumImageSize;
        }

        if (bitmap.Width > MaximumImageSize)
        {
            bitmapImage.DecodePixelWidth = MaximumImageSize;
        }

        bitmapImage.EndInit();
        bitmapImage.Freeze();

        return bitmapImage;
    }

    private static BitmapSource CreateBitmapSourceInternal(Bitmap bitmap)
    {
        var bitmapData = bitmap.LockBits(
                new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                ImageLockMode.ReadWrite,
                StandartPixelFormat);

        try
        {
            const int BufferModifier = 4;

            var bitmapSource = BitmapSource.Create(
                bitmap.Width,
                bitmap.Height,
                bitmap.HorizontalResolution,
                bitmap.VerticalResolution,
                PixelFormats.Bgra32,
                palette: null,
                bitmapData.Scan0,
                bitmap.Width * bitmap.Height * BufferModifier,
                bitmapData.Stride);

            bitmapSource.Freeze();
            return bitmapSource;
        }
        finally
        {
            bitmap.UnlockBits(bitmapData);
        }
    }
}
